"""
smoke_test.py
"""

from .decode import decode

# The format for tests is a dict of { "test_name" : value to pass to threefive.decode}
ten_tests = {
    "Base64": "/DAvAAAAAAAA///wBQb+dGKQoAAZAhdDVUVJSAAAjn+fCAgAAAAALKChijUCAKnMZ1g=",
    "Bytes": b"\xfc0\x11\x00\x00\x00\x00\x00\x00\x00\xff\xff\xff\x00\x00\x00O%3\x96",
    "Hex String": "0XFC301100000000000000FFFFFF0000004F253396",
    "Hex Literal": 0xFC301100000000000000FFFFFF0000004F253396,
    "Integer": 1439737590925997869941740173214217318917816529814,
    "HTTP/HTTPS Streams": "https://futzu.com/xaa.ts",
    # "Bad" tests are expected to fail.
    "Bad Base64 ": "/DAvAf45AA",
    "Bad File": "/you/me/fake.file",
    "Bad Integer": -0.345,
    " Bad String": "your momma",
}


def _decode_test(test_name, test_data):
    passed = "✔"
    failed = "✘"
    print(f"testing {test_name}\n Data: {test_data}\n")
    if decode(test_data):
        return passed
    return failed


def smoke(tests=None):
    """
    calls threefive.decode using the values in tests.
    The format for tests:
    { "test_name" : value to pass to threefive.decode}

    example:

     my_tests ={
    "Base64": "/DAvAAAAAAAA///wBQb+dGKQoAAZAhdDVUVJSAAAjn+fCAgAAAAALKChijUCAKnMZ1g=",
    "Bytes": b"\xfc0\x11\x00\x00\x00\x00\x00\x00\x00\xff\xff\xff\x00\x00\x00O%3\x96"
    }

    import threefive
    threefive.smoke_test(my_tests)

    """
    if not tests:
        tests = ten_tests
    results = {k: _decode_test(k, v) for k, v in tests.items()}
    print("Smoke Test\n")
    for kay, vee in results.items():
        print(f"{kay}  {vee}")


if __name__ == "__main__":
    smoke()



"""
threefive.__init__.py
"""

from .cue import Cue
from .decode import decode
from .section import SpliceInfoSection
from .segment import Segment
from .smoketest import smoke
from .stream import Stream
from .version import version, version_number

from .commands import (
    TimeSignal,
    SpliceInsert,
    SpliceNull,
    SpliceSchedule,
    PrivateCommand,
    BandwidthReservation,
)

from .descriptors import (
    AudioDescriptor,
    AvailDescriptor,
    DtmfDescriptor,
    SegmentationDescriptor,
    TimeDescriptor,
)

from .encode import (
    mk_splice_null,
    mk_splice_insert,
    mk_time_signal,
)



"""
decode.py

decode is a SCTE-35 decoder function
with input type auto-detection.

SCTE-35 data can be parsed with just
one function call.

the arg stuff is the input.
if stuff is not set, decode will attempt
to read mpegts video from sys.stdin.buffer.

SCTE-35 data is printed in JSON format.

For more parsing and output control,
see the Cue and Stream classes.

"""

import sys

from .cue import Cue
from .stream import Stream


def _read_stuff(stuff):
    try:
        # Mpegts Video
        strm = Stream(stuff)
        strm.decode()
        return True
    except:
        try:
            cue = Cue(stuff)
            cue.decode()
            cue.show()
            return True
        except:
            return False


def decode(stuff=None):
    """
    decode is a SCTE-35 decoder function
    with input type auto-detection.

    SCTE-35 data is printed in JSON format.

    Use like:

    # Base64
    stuff = '/DAvAAAAAAAA///wBQb+dGKQoAAZAhdDVUVJSAAAjn+fCAgAAAAALKChijUCAKnMZ1g='
    threefive.decode(stuff)

    # Bytes
    payload = b"\xfc0\x11\x00\x00\x00\x00\x00\x00\x00\xff\xff\xff\x00\x00\x00O%3\x96"
    threefive.decode(payload)

    # Hex String
    stuff = '0XFC301100000000000000FFFFFF0000004F253396'
    threefive.decode(stuff)

    # Hex Literal
    threefive.decode(0XFC301100000000000000FFFFFF0000004F253396)

    # Integer
    big_int = 1439737590925997869941740173214217318917816529814
    threefive.decode(big_int)

    # Mpegts File
    threefive.decode('/path/to/mpegts')

    # Mpegts HTTP/HTTPS Streams
    threefive.decode('https://futzu.com/xaa.ts')

    """
    if stuff in [None, sys.stdin.buffer]:
        # Mpegts stream or file piped in
        return Stream(sys.stdin.buffer).decode()
    if isinstance(stuff, int):
        return _read_stuff(hex(stuff))
    return _read_stuff(stuff)



"""
The threefive.Segment class
"""
import os
from new_reader import reader
from .stream import Stream

AES = True
try:
    import pyaes
except:
    AES = False


class Segment(Stream):
    """
    The Segment class is a Sub Class of threefive.Stream
    made for small, fixed size MPEGTS files,like HLS segments.

    Segment Class Specific Features:

    * Decryption of AES Encrypted MPEGTS.

    * Segment.cues  a list of SCTE35 cues found in the segment.


    Example:

        from threefive import Segment

        >>>> uri = "https://example.com/1.ts"
        >>>> seg = Segment(uri)
        >>>> seg.decode()
        >>>> [cue.encode() for cue in cues]
        ['/DARAAAAAAAAAP/wAAAAAHpPv/8=',
        '/DAvAAAAAAAAAP/wFAUAAAKWf+//4WoauH4BTFYgAAEAAAAKAAhDVUVJAAAAAOv1oqc=']

        # For aes encrypted files

        >>>> key = "https://example.com/aes.key"
        >>>> IV=0x998C575D24F514AEC84EDC5CABCCDB81
        >>>> uri = "https://example.com/aes-1.ts"

        >>>> seg = Segment(uri,key_uri=key, iv=IV)
        >>>> seg.decode()
        >>>> {cue.packet_data.pcr:cue.encode() for cue in seg.cues}

       { 89718.451333: '/DARAAAAAAAAAP/wAAAAAHpPv/8=',
       89730.281789: '/DAvAAAAAAAAAP/wFAUAAAKWf+//4WoauH4BTFYgAAEAAAAKAAhDVUVJAAAAAOv1oqc='}

    """

    def __init__(self, seg_uri, key_uri=None, iv=None):
        self.seg_uri = seg_uri
        self.key_uri = key_uri
        self.key = None
        self.iv = None
        self.cues = []
        self.pts_start = None
        self.tmp = None
        if AES:
            if iv:
                self.iv = int.to_bytes(int(iv), 16, byteorder="big")
            if self.key_uri:
                self._aes_get_key()
                self._aes_decrypt()
        super().__init__(self.seg_uri)

    def __repr__(self):
        return str(self.__dict__)

    def _mk_tmp(self):
        self.tmp = "tf-"
        self.tmp += self.seg_uri.rsplit("/", 1)[-1]

    def _aes_get_key(self):
        with reader(self.key_uri) as quay:
            self.key = quay.read()

    def _aes_decrypt(self):
        mode = pyaes.AESModeOfOperationCBC(self.key, iv=self.iv)
        self._mk_tmp()
        with open(self.tmp, "wb") as outfile, reader(self.seg_uri) as infile:
            pyaes.decrypt_stream(mode, infile, outfile)
        self.seg_uri = self.tmp

    def add_cue(self, cue):
        """
        add_cue is called  by a segment instance
        to collect SCTE35 cues.
        """
        self.cues.append(cue)

    def show_cue(self, cue):
        """
        show_cue prints SCTE35 Cue data
        and calls add_cue to append the cue to
        the Segment,cues list.
        """
        cue.show()
        self.add_cue(cue)

    def decode(self, func=None):
        """
        decode a mpegts segment.
        """
        super().decode(func=self.show_cue)
        # self.pts_start

        if self.tmp:
            os.unlink(self.tmp)



"""
threefive.Cue Class
"""
from base64 import b64decode, b64encode
import json
from sys import stderr
from .bitn import NBin
from .base import SCTE35Base
from .section import SpliceInfoSection
from .commands import command_map
from .descriptors import splice_descriptor, descriptor_map
from .crc import crc32


class Cue(SCTE35Base):
    """
    The threefive.Splice class handles parsing
    SCTE 35 message strings.
    Example usage:

    >>>> import threefive
    >>>> Base64 = "/DAvAAAAAAAA///wBQb+dGKQoAAZAhdDVUVJSAAAjn+fCAgAAAAALKChijUCAKnMZ1g="
    >>>> cue = threefive.Cue(Base64)

    # cue.decode() returns True on success,or False if decoding failed

    >>>> cue.decode()
    True

    # After Calling cue.decode() the instance variables can be accessed via dot notation.

    >>>> cue.command
    {'command_length': 5, 'name': 'Time Signal', 'time_specified_flag': True,
    'pts_time': 21695.740089}

    >>>> cue.command.pts_time
    21695.740089

    >>>> cue.info_section.table_id

    '0xfc'

    """

    def __init__(self, data=None, packet_data=None):
        """
        data may be packet bites or encoded string
        packet_data is a instance passed from a Stream instance
        """
        self.bites = None
        if data:
            self.bites = self._mk_bits(data)
        self.info_section = SpliceInfoSection()
        self.command = None
        self.descriptors = []
        self.packet_data = packet_data

    def __repr__(self):
        return str(self.__dict__)

    # decode

    def decode(self):
        """
        Cue.decode() parses for SCTE35 data
        """
        bites = self.bites
        self.descriptors = []
        while bites:
            bites = self.mk_info_section(bites)
            bites = self._set_splice_command(bites)
            bites = self._mk_descriptors(bites)
            crc = hex(int.from_bytes(bites[0:4], byteorder="big"))
            self.info_section.crc = crc
            return True
        return False

    def _descriptor_loop(self, loop_bites):
        """
        Cue._descriptor_loop parses all splice descriptors
        """
        tag_n_len = 2
        while loop_bites:
            spliced = splice_descriptor(loop_bites)
            if not spliced:
                return
            sd_size = tag_n_len + spliced.descriptor_length
            loop_bites = loop_bites[sd_size:]
            del spliced.bites
            self.descriptors.append(spliced)

    def get(self):
        """
        Cue.get returns the SCTE-35 Cue
        data as a dict of dicts.
        """
        if self.command and self.info_section:
            scte35 = {
                "info_section": self.info_section.get(),
                "command": self.command.get(),
                "descriptors": self.get_descriptors(),
            }
            if self.packet_data is not None:
                scte35["packet_data"] = self.packet_data.get()
            return scte35
        return False

    def get_descriptors(self):
        """
        Cue.get_descriptors returns a list of
        SCTE 35 splice descriptors as dicts.
        """
        return [d.get() for d in self.descriptors]

    def get_json(self):
        """
        Cue.get_json returns the Cue instance
        data in json.
        """
        return json.dumps(self.get(), indent=4)

    @staticmethod
    def _mk_bits(data):
        """
        cue._mk_bits Converts
        Hex and Base64 strings into bytes.
        """
        if isinstance(data, bytes):
            return data[data.index(b"\xfc") :]
        # handles int and unquoted hex
        if isinstance(data, int):
            length = data.bit_length() >> 3
            bites = int.to_bytes(data, length, byteorder="big")
            return bites
        try:
            # Handles hex byte strings
            i = int(data, 16)
            i_len = i.bit_length() >> 3
            bites = int.to_bytes(i, i_len, byteorder="big")
            return bites
        except (LookupError, TypeError, ValueError):
            if data[:2].lower() == "0x":
                data = data[2:]
            if data[:2].lower() == "fc":
                return bytes.fromhex(data)
        try:
            return b64decode(data)
        except (LookupError, TypeError, ValueError):
            return data

    def _mk_descriptors(self, bites):
        """
        Cue._mk_descriptors parses
        Cue.info_section.descriptor_loop_length,
        then call Cue._descriptor_loop
        """
        dll = (bites[0] << 8) | bites[1]
        self.info_section.descriptor_loop_length = dll
        bites = bites[2:]
        self._descriptor_loop(bites[:dll])
        return bites[dll:]

    def mk_info_section(self, bites):
        """
        Cue.mk_info_section parses the
        Splice Info Section
        of a SCTE35 cue.
        """
        info_size = 14
        info_bites = bites[:info_size]
        self.info_section.decode(info_bites)
        return bites[info_size:]

    def _set_splice_command(self, bites):
        """
        Cue._set_splice_command parses
        the command section of a SCTE35 cue.
        """
        sct = self.info_section.splice_command_type
        if sct not in command_map:
            return False
        self.command = command_map[sct](bites)
        self.command.decode()
        del self.command.bites
        self.info_section.splice_command_length = self.command.command_length
        bites = bites[self.command.command_length :]
        return bites

    def show(self):
        """
        Cue.show prints the Cue as JSON
        """
        print(self.get_json())

    def to_stderr(self):
        """
        Cue.to_stderr prints the Cue
        as JSON to sys.stderr
        """
        print(self.get_json(), file=stderr)

    # encode related

    def encode(self):
        """
        Cue.encode() converts SCTE35 data
        to a base64 encoded string.
        """
        dscptr_bites = self._unloop_descriptors()
        dll = len(dscptr_bites)
        self.info_section.descriptor_loop_length = dll
        if not self.command:
            err_mesg = "\033[7mA splice command is required\033[27m"
            raise ValueError(err_mesg)
        cmd_bites = self.command.encode()
        cmdl = self.command.command_length = len(cmd_bites)
        self.info_section.splice_command_length = cmdl
        self.info_section.splice_command_type = self.command.command_type
        # 11 bytes for info section + command + 2 descriptor loop length
        # + descriptor loop + 4 for crc
        self.info_section.section_length = 11 + cmdl + 2 + dll + 4
        self.bites = self.info_section.encode()
        self.bites += cmd_bites
        self.bites += int.to_bytes(
            self.info_section.descriptor_loop_length, 2, byteorder="big"
        )
        self.bites += dscptr_bites
        self._encode_crc()
        return b64encode(self.bites).decode()

    def encode_as_int(self):
        """
        encode_as_int returns self.bites as an int.
        """
        self.encode()
        return int.from_bytes(self.bites, byteorder="big")

    def encode_as_hex(self):
        """
        encode_as_hex returns self.bites as
        a hex string
        """
        return hex(self.encode_as_int())

    def _encode_crc(self):
        """
        _encode_crc encode crc32
        """
        crc_int = crc32(self.bites)
        self.info_section.crc = hex(crc_int)
        self.bites += int.to_bytes(crc_int, 4, byteorder="big")

    def _unloop_descriptors(self):
        """
        _unloop_descriptors
        for each descriptor in self.descriptors
        encode descriptor tag, descriptor length,
        and the descriptor into all_bites.bites
        """
        all_bites = NBin()
        dbite_chunks = [dsptr.encode() for dsptr in self.descriptors]
        for chunk, dsptr in zip(dbite_chunks, self.descriptors):
            dsptr.descriptor_length = len(chunk)
            all_bites.add_int(dsptr.tag, 8)
            all_bites.add_int(dsptr.descriptor_length, 8)
            all_bites.add_bites(chunk)
        return all_bites.bites

    def load(self, stuff):
        """
        Cue.load loads SCTE35 data for encoding.
        stuff is a dict or json
        with any or all of these keys
        stuff = {
            'info_section': {dict} ,
            'command': {dict},
            'descriptors': [list of {dicts}],
            }
        """
        if isinstance(stuff, str):
            stuff = json.loads(stuff)
        if "info_section" in stuff:
            self.load_info_section(stuff["info_section"])
        if "command" in stuff:
            self.load_command(stuff["command"])
        if "descriptors" in stuff:
            self.load_descriptors(stuff["descriptors"])

    def load_info_section(self, isec):
        """
        load_info_section loads data for Cue.info_section
        isec should be a dict.
        if 'splice_command_type' is included,
        an empty command instance will be created for Cue.command
        """
        self.info_section.load(isec)
        if "splice_command_type" in isec:
            cmd_type = isec["splice_command_type"]
            command_map[cmd_type]()

    def load_command(self, cmd):
        """
        load_command loads data for Cue.command
        cmd should be a dict.
        if 'command_type' is included,
        the command instance will be created.
        """
        if not isinstance(cmd, dict):
            try:
                cmd = cmd.get()
            except:
                print(f" cmd is a {type(cmd)} should be a dict,")
                return
        if "command_type" in cmd:
            self.command = command_map[cmd["command_type"]]()
        if self.command:
            self.command.load(cmd)

    def load_descriptors(self, dlist):
        """
        Load_descriptors loads descriptor data.
        dlist is a list of dicts
        if 'tag' is included in each dict,
        a descriptor instance will be created.
        """
        if not isinstance(dlist, list):
            raise Exception("descriptors should be a list")
        for dstuff in dlist:
            if not isinstance(dstuff, dict):
                try:
                    dstuff = dstuff.get()
                except:
                    print(f" dstuff is a {type(dstuff)} should be a dict,")
                    return
            if "tag" in dstuff:
                dscptr = descriptor_map[dstuff["tag"]]()
                dscptr.load(dstuff)
                self.descriptors.append(dscptr)



"""
packetdata.py
"""

from .base import SCTE35Base


class PacketData(SCTE35Base):
    """
    A PacketData instance is used
    to hold packet pid, program, pcr and pts
    for SCTE-35 packets.
    """

    def __init__(self, pid, prgm):
        self.pid = hex(pid)
        self.program = prgm
        self.pcr_ticks = None
        self.pcr = None
        self.pts_ticks = None
        self.pts = None

    @staticmethod
    def _mk_timestamp(seconds):
        if seconds:
            return round((seconds / 90000.0), 6)
        return seconds

    def mk_pcr(self, table):
        """
        calculates and formats pcr
        """
        try:
            self.pcr_ticks = table[self.program]
            self.pcr = self._mk_timestamp(self.pcr_ticks)
        except:
            pass

    def mk_pts(self, table):
        """
        mk_pts calculates and formats pts
        """
        try:
            self.pts_ticks = table[self.program]
            self.pts = self._mk_timestamp(self.pts_ticks)
        except:
            pass



"""
encode.py

threefive.encode has helper functions for Cue encoding.

"""


from .commands import SpliceNull, SpliceInsert, TimeSignal
from .cue import Cue


def mk_splice_null():
    """
    mk_splice_null returns a Cue
    with a Splice Null
    """
    cue = Cue()
    sn = SpliceNull()
    cue.command = sn
    cue.encode()
    return cue


def mk_time_signal(pts=None):
    """
     mk_time_signal returns a Cue
     with a Time Signal

     if pts is NOT set:
         time_specified_flag   False

    if pts IS set:
         time_specified_flag   True
         pts_time              pts

    """
    cue = Cue()
    ts = TimeSignal()
    ts.time_specified_flag = False
    if pts:
        pts = float(pts)
        ts.time_specified_flag = True
        ts.pts_time = pts
    cue.command = ts
    cue.encode()
    return cue


def mk_splice_insert(event_id, pts, duration=None):
    """
    mk_cue returns a Cue
    with a Splice Insert.

    splice_event_id = event_id

    If duration is NOT set,
        out_of_network_indicator   False
        time_specified_flag        False
        duration_flag              False
        splice_immediate_flag      True

    if duration IS set:
        out_of_network_indicator   True
        time_specified_flag        True
        duration_flag              True
        splice_immediate_flag      False
        break_auto_return          True
        break_duration             duration
        pts_time                   pts

    """
    pts = float(pts)
    cue = Cue()
    # default is a CUE-IN
    sin = SpliceInsert()
    sin.splice_event_id = event_id
    sin.splice_event_cancel_indicator = False
    sin.out_of_network_indicator = False
    sin.time_specified_flag = False
    sin.program_splice_flag = True
    sin.duration_flag = False
    sin.splice_immediate_flag = True
    sin.unique_program_id = event_id
    sin.avail_num = 0
    sin.avail_expected = 0
    # If we have a duration, make a CUE-OUT
    if duration is not None:
        duration = float(duration)
        sin.time_specified_flag = True
        sin.break_duration = duration
        sin.break_auto_return = True
        sin.splice_immediate_flag = False
        sin.duration_flag = True
        sin.out_of_network_indicator = True
        sin.pts_time = pts
    cue.command = sin  # Add SpliceInsert to the SCTE35 cue
    cue.encode()
    return cue



"""
Mpeg-TS Stream parsing class Stream
"""
import sys
from functools import partial
from new_reader import reader
from .cue import Cue
from .packetdata import PacketData

"""
stream types for program streams.
"""
streamtype_map = {
    "0x2": "MP2 Video",
    "0x3": "MP2 Audio",
    "0x4": "MP2 Audio",
    "0x6": "PES Packets/Private Data",
    "0xf": "AAC Audio",
    "0x1b": "AVC Video",
    "0x81": "AC3 Audio ",
    "0x86": "SCTE35 Data",
    "0xc0": "Unknown",
}


def no_op(cue):
    """
    no_op is just a dummy func to pass to Stream.decode()
    to suppress output.
    """
    return cue


def show_cue(cue):
    """
    default function call for Stream.decode
    when a SCTE-35 packet is found.
    """
    cue.show()


def show_cue_stderr(cue):
    """
    print cue data to sys.stderr
    for Stream.decode_proxy
    """
    cue.to_stderr()


class ProgramInfo:
    """
    ProgramInfo is a class to
    hold Program information
    for use with Stream.show()
    """

    def __init__(self, pid=None, pcr_pid=None):
        self.pid = pid
        self.pcr_pid = pcr_pid
        self.provider = b""
        self.service = b""
        self.streams = {}  # pid to stream_type mapping

    def show(self):
        """
        show print the Program Infomation
        in a familiar format.
        """
        serv = self.service.decode(errors="ignore")
        prov = self.provider.decode(errors="ignore")
        print(f"    Service:\t{ serv}\n    Provider:\t{prov}")
        print(f"    Pid:\t{self.pid}")
        print(f"    Pcr Pid:\t{self.pcr_pid}")
        print("    Streams:")
        # sorted_dict = {k:my_dict[k] for k in sorted(my_dict)})
        keys = sorted(self.streams)
        for k in keys:
            vee = self.streams[k]
            if vee in streamtype_map:
                vee = f"{vee} {streamtype_map[vee]}"
            else:
                vee = f"{vee} Unknown"
            print(f"\t\tPid: {k}[{hex(k)}]\tType: {vee}")
        print()


class Pids:
    SDT_PID = 0x11
    PAT_PID = 0x00

    def __init__(self):
        self.pcr = set()
        self.pmt = set()
        self.scte35 = set()
        self.tables = set()
        self.tables.add(self.PAT_PID)
        self.tables.add(self.SDT_PID)


class Maps:
    def __init__(self):
        self.pid_cc = {}
        self.pid_prgm = {}
        self.prgm_pcr = {}
        self.prgm_pts = {}
        self.prgm = {}
        self.partial = {}
        self.last = {}


class Stream:
    """
    Stream class for parsing MPEG-TS data.
    """

    _PACKET_SIZE = 188
    _SYNC_BYTE = 0x47
    # tids
    _PMT_TID = b"\x02"
    _SCTE35_TID = b"\xFC0"
    _SDT_TID = b"\x42"
    # pts
    ROLLOVER = 8589934591  # 95443.717678
    _NO_PTS_PIDS = [188, 190, 191, 240, 241, 242, 248]

    def __init__(self, tsdata, show_null=True):
        """
        tsdata is an file or http/https url
        set show_null=False to exclude Splice Nulls

        Use like...

        from threefive import Stream
        strm = Stream("vid.ts",show_null=False)
        strm.decode()

        """
        if isinstance(tsdata, str):
            self._tsdata = reader(tsdata)
        else:
            self._tsdata = tsdata
        self.show_null = show_null
        self.start = {}
        self.info = None
        self.the_program = None
        self.pids = Pids()
        self.maps = Maps()

    def __repr__(self):
        return str(self.__dict__)

    def _find_start(self):
        while self._tsdata:
            one = self._tsdata.read(1)
            if not one:
                print("\nNo Stream Found. \n")
                return False
            if one[0] == self._SYNC_BYTE:
                tail = self._tsdata.read(self._PACKET_SIZE - 1)
                if tail:
                    self._parse(one + tail)
                    return True

    def decode(self, func=show_cue):
        """
        Stream.decode reads self.tsdata to find SCTE35 packets.
        func can be set to a custom function that accepts
        a threefive.Cue instance as it's only argument.
        """
        if not self._find_start():
            return False
        for pkt in iter(partial(self._tsdata.read, self._PACKET_SIZE), b""):
            cue = self._parse(pkt)
            if cue:
                if not func:
                    return cue
                func(cue)
        self._tsdata.close()
        return True

    def decode2(self, func=show_cue):
        """
        Stream.decode2 reads self.tsdata to find SCTE35 packets.
        func can be set to a custom function that accepts
        a threefive.Cue instance as it's only argument.

        Stream.decode2 is super fast with pypy3

        """
        if not self._find_start():
            return
        with self._tsdata as data:
            while data:
                pkt = data.read(self._PACKET_SIZE)
                if not pkt:
                    break
                cue = self._parse(pkt)
                if cue:
                    func(cue)

    def decode_next(self):
        """
        Stream.decode_next returns the next
        SCTE35 cue as a threefive.Cue instance.
        """
        cue = self.decode(func=False)
        if cue:
            return cue
        return False

    def decode_program(self, the_program, func=show_cue):
        """
        Stream.decode_program limits SCTE35 parsing
        to a specific MPEGTS program.
        """
        self.the_program = the_program
        return self.decode(func)

    def decode_proxy(self, func=show_cue_stderr):
        """
        Stream.decode_proxy writes all ts packets are written to stdout
        for piping into another program like mplayer.
        SCTE-35 cues are printed to stderr.
        """
        if self._find_start():
            for pkt in iter(partial(self._tsdata.read, self._PACKET_SIZE), b""):
                cue = self._parse(pkt)
                if cue:
                    func(cue)
                sys.stdout.buffer.write(pkt)
            self._tsdata.close()

    def show(self):
        """
        displays streams that will be
        parsed for SCTE-35.
        """
        self.info = True
        if self._find_start():
            for pkt in iter(partial(self._tsdata.read, self._PACKET_SIZE), b""):
                self._parse_info(pkt)
            sopro = sorted(self.maps.prgm.items())
            for k, vee in sopro:
                if len(vee.streams.items()) > 0:
                    print(f"Program: {k}")
                    vee.show()
        return True

    def decode_start_time(self):
        """
        displays streams that will be
        parsed for SCTE-35.
        """
        self.decode(func=no_op)
        if len(self.start.values()) > 0:
            return self.start.popitem()[1]
        return False

    def _mk_packet_data(self, pid):
        prgm = self.maps.pid_prgm[pid]
        pdata = PacketData(pid, prgm)
        pdata.mk_pcr(self.maps.prgm_pcr)
        pdata.mk_pts(self.maps.prgm_pts)
        return pdata

    @staticmethod
    def _afc_flag(pkt):
        return pkt[3] & 0x20

    @staticmethod
    def _pcr_flag(pkt):
        return pkt[5] & 0x10

    @staticmethod
    def _pts_flag(pay):
        # uses pay not pkt
        return pay[7] & 0x80

    @staticmethod
    def _pusi_flag(pkt):
        return pkt[1] & 0x40

    @staticmethod
    def _spi_flag(pkt):
        return pkt[5] & 0x20

    @staticmethod
    def _parse_length(byte1, byte2):
        """
        parse a 12 bit length value
        """
        return (byte1 & 0x0F) << 8 | byte2

    @staticmethod
    def _parse_pid(byte1, byte2):
        """
        parse a 13 bit pid value
        """
        return (byte1 & 0x01F) << 8 | byte2

    @staticmethod
    def _parse_program(byte1, byte2):
        """
        parse a 16 bit program number value
        """
        return (byte1 << 8) | byte2

    @staticmethod
    def _split_by_idx(pay, marker):
        try:
            return pay[pay.index(marker) :]
        except (LookupError, TypeError, ValueError):
            return False

    def _parse_pts(self, pkt, pid):
        """
        parse pts and store by program key
        in the dict Stream._pid_pts
        """
        payload = self._parse_payload(pkt)
        if len(payload) < 14:
            return
        if self._pts_flag(payload):
            pts = ((payload[9] >> 1) & 7) << 30
            pts |= payload[10] << 22
            pts |= (payload[11] >> 1) << 15
            pts |= payload[12] << 7
            pts |= payload[13] >> 1
            prgm = self.pid2prgm(pid)
            self.maps.prgm_pts[prgm] = pts
            if prgm not in self.start:
                self.start[prgm] = pts

    def _parse_pcr(self, pkt, pid):
        """
        parse pcr and store by program key
        in the dict Stream._pid_pcr
        """
        if self._afc_flag(pkt):
            if self._pcr_flag(pkt):
                pcr = pkt[6] << 25
                pcr |= pkt[7] << 17
                pcr |= pkt[8] << 9
                pcr |= pkt[9] << 1
                pcr |= pkt[10] >> 7
                prgm = 1
                if pid in self.maps.pid_prgm:
                    prgm = self.maps.pid_prgm[pid]
                self.maps.prgm_pcr[prgm] = pcr

    def _parse_cc(self, pkt, pid):
        last_cc = None
        c_c = pkt[3] & 0xF
        if pid in self.maps.pid_cc:
            last_cc = self.maps.pid_cc[pid]
            good = (last_cc, last_cc + 1)
            if last_cc == 15:
                good = (15, 0)
            if c_c not in good:
                print(
                    f"BAD --> pid: {hex(pid)} last cc: {last_cc} cc: {c_c}",
                    file=sys.stderr,
                )
            self.maps.pid_cc[pid] = c_c

    @staticmethod
    def as_90k(ticks):
        """
        as_90k returns ticks as 90k clock time
        """
        return round((ticks / 90000.0), 6)

    def pid2prgm(self, pid):
        """
        pid2prgm takes a pid,
        returns the program
        """
        prgm = 1
        if pid in self.maps.pid_prgm:
            prgm = self.maps.pid_prgm[pid]
        return prgm

    def pid2pts(self, pid):
        """
        pid2pts takes a pid
        returns the current pts
        """
        prgm = self.pid2prgm(pid)
        if prgm not in self.maps.prgm_pts:
            return False
        return self.as_90k(self.maps.prgm_pts[prgm])

    def pid2pcr(self, pid):
        """
        pid2pcr takes a pid
        returns the current pcr
        """
        prgm = self.pid2prgm(pid)
        if prgm not in self.maps.prgm_pcr:
            return False
        return self.as_90k(self.maps.prgm_pcr[prgm])

    def _parse_payload(self, pkt):
        head_size = 4
        if self._afc_flag(pkt):
            afl = pkt[4]
            head_size += afl + 1  # +1 for afl byte
        return pkt[head_size:]

    def _parse_tables(self, pkt, pid):
        """
        _parse_tables parse for
        PAT, PMT,  and SDT tables
        based on pid of the pkt
        """
        pay = self._parse_payload(pkt)
        if not self._chk_last(pay, pid):
            if pid == self.pids.PAT_PID:
                return self._program_association_table(pay)
            if pid == self.pids.SDT_PID:
                if self.info:
                    return self._stream_descriptor_table(pay)
            if pid in self.pids.pmt:
                return self._program_map_table(pay, pid)
        return False

    def _parse_info(self, pkt):
        pid = self._parse_pid(pkt[1], pkt[2])
        if pid in self.pids.tables:
            self._parse_tables(pkt, pid)
        return pid

    def _parse(self, pkt):
        cue = False
        pid = self._parse_pid(pkt[1], pkt[2])
        if pid in self.pids.tables:
            self._parse_tables(pkt, pid)
        if pid in self.pids.pcr:
            self._parse_cc(pkt, pid)
            self._parse_pcr(pkt, pid)
        if self._pusi_flag(pkt):
            self._parse_pts(pkt, pid)
        if pid in self.pids.scte35:
            cue = self._parse_scte35(pkt, pid)
        return cue

    def _chk_partial(self, pay, pid, sep):
        if pid in self.maps.partial:
            pay = self.maps.partial.pop(pid) + pay
        return self._split_by_idx(pay, sep)

    def _chk_last(self, pay, pid):
        if pid in self.maps.last:
            return pay == self.maps.last[pid]
        self.maps.last[pid] = pay
        return False

    def _section_done(self, pay, pid, seclen):
        # + 3 for the bytes before section starts
        if (seclen + 3) > len(pay):
            self.maps.partial[pid] = pay
            return False
        return True

    def _parse_cue(self, pay, pid):
        packet_data = None
        packet_data = self._mk_packet_data(pid)
        cue = Cue(pay, packet_data)
        if cue.decode():
            return cue
        return False

    def _parse_scte35(self, pkt, pid):
        """
        parse a scte35 cue from one or more packets
        """
        pay = self._parse_payload(pkt)
        if not pay:
            return False
        pay = self._chk_partial(pay, pid, self._SCTE35_TID)
        if not pay:
            self.pids.scte35.remove(pid)
            return False
        if pay[13] == self.show_null:
            return False
        seclen = self._parse_length(pay[1], pay[2])
        if not self._section_done(pay, pid, seclen):
            return False
        pay = pay[: seclen + 3]
        cue = self._parse_cue(pay, pid)
        return cue

    def _stream_descriptor_table(self, pay):
        """
        _stream_descriptor_table parses the SDT for program metadata
        """
        pay = self._chk_partial(pay, self.pids.SDT_PID, self._SDT_TID)
        if not pay:
            return False
        seclen = self._parse_length(pay[1], pay[2])
        if not self._section_done(pay, self.pids.SDT_PID, seclen):
            return False
        idx = 11
        while idx < seclen + 3:
            service_id = self._parse_program(pay[idx], pay[idx + 1])
            idx += 3
            dloop_len = self._parse_length(pay[idx], pay[idx + 1])
            idx += 2
            i = 0
            while i < dloop_len:
                if pay[idx] == 0x48:
                    i += 3
                    spnl = pay[idx + i]
                    i += 1
                    service_provider_name = pay[idx + i : idx + i + spnl]
                    i += spnl
                    snl = pay[idx + i]
                    i += 1
                    service_name = pay[idx + i : idx + i + snl]
                    i += snl
                    if service_id not in self.maps.prgm:
                        self.maps.prgm[service_id] = ProgramInfo()
                    pinfo = self.maps.prgm[service_id]
                    pinfo.provider = service_provider_name
                    pinfo.service = service_name
                i = dloop_len
                idx += i

    def _program_association_table(self, pay):
        """
        parse program association table
        for program to pmt_pid mappings.
        """
        pay = self._chk_partial(pay, self.pids.PAT_PID, b"")
        seclen = self._parse_length(pay[2], pay[3])
        if not self._section_done(pay, self.pids.PAT_PID, seclen):
            return False
        seclen -= 5  # pay bytes 4,5,6,7,8
        idx = 9
        chunk_size = 4
        while seclen > 4:  #  4 bytes for crc
            program_number = self._parse_program(pay[idx], pay[idx + 1])
            if program_number > 0:
                pmt_pid = self._parse_pid(pay[idx + 2], pay[idx + 3])
                self.pids.pmt.add(pmt_pid)
                self.pids.tables.add(pmt_pid)

            seclen -= chunk_size
            idx += chunk_size

    def _program_map_table(self, pay, pid):
        """
        parse program maps for streams
        """
        pay = self._chk_partial(pay, pid, self._PMT_TID)
        if not pay:
            return False
        seclen = self._parse_length(pay[1], pay[2])
        if not self._section_done(pay, pid, seclen):
            return False
        program_number = self._parse_program(pay[3], pay[4])
        if self.the_program and (program_number != self.the_program):
            return False
        pcr_pid = self._parse_pid(pay[8], pay[9])
        if program_number not in self.maps.prgm:
            self.maps.prgm[program_number] = ProgramInfo()
        pinfo = self.maps.prgm[program_number]
        pinfo.pid = pid
        pinfo.pcr_pid = pcr_pid
        self.pids.pcr.add(pcr_pid)
        self.maps.pid_prgm[pcr_pid] = program_number
        proginfolen = self._parse_length(pay[10], pay[11])
        idx = 12
        idx += proginfolen
        si_len = seclen - 9
        si_len -= proginfolen
        self._parse_program_streams(si_len, pay, idx, program_number)
        return True

    def _parse_program_streams(self, si_len, pay, idx, program_number):
        """
        parse the elementary streams
        from a program
        """
        # 5 bytes for stream_type info
        chunk_size = 5
        end_idx = (idx + si_len) - chunk_size
        while idx < end_idx:
            stream_type, pid, ei_len = self._parse_stream_type(pay, idx)
            pinfo = self.maps.prgm[program_number]
            pinfo.streams[pid] = stream_type
            idx += chunk_size
            idx += ei_len
            self.maps.pid_prgm[pid] = program_number
            self._chk_pid_stream_type(pid, stream_type)

    def _parse_stream_type(self, pay, idx):
        """
        extract stream pid and type
        """
        stream_type = hex(pay[idx])
        el_pid = self._parse_pid(pay[idx + 1], pay[idx + 2])
        ei_len = self._parse_length(pay[idx + 3], pay[idx + 4])
        return stream_type, el_pid, ei_len

    def _chk_pid_stream_type(self, pid, stream_type):
        """
        if stream_type is 0x06 or 0x86
        add it to self._scte35_pids.
        """
        if stream_type in ["0x6", "0x86"]:
            self.pids.scte35.add(pid)



"""
threefive/upids.py

threefve.upids

"""
charset = "ascii"


class UpidDecoder:
    """
    UpidDecoder for use by the
    SegmentationDescriptor class.
    """

    def __init__(self, bitbin, upid_type, upid_length):
        self.bitbin = bitbin
        self.upid_type = upid_type
        self.upid_length = upid_length
        self.bit_length = upid_length << 3

    def _decode_air_id(self):
        return self.bitbin.as_hex(self.bit_length)

    def _decode_atsc(self):
        cont_size = self.bit_length - 32
        return {
            "TSID": self.bitbin.as_int(16),
            "reserved": self.bitbin.as_int(2),
            "end_of_day": self.bitbin.as_int(5),
            "unique_for": self.bitbin.as_int(9),
            "content_id": self.bitbin.as_bytes(cont_size),
        }

    def _decode_eidr(self):
        if self.upid_length != 12:
            return f"upid_length is {self.upid_length} should be 12 bytes."
        pre = self.bitbin.as_int(16)
        post = []
        bit_count = 80
        while bit_count:
            bit_count -= 16
            post.append(self.bitbin.as_hex(16)[2:])
        return f"10.{pre}/{'-'.join(post)}"

    def _decode_isan(self):
        return self.bitbin.as_hex(self.bit_length)

    def _decode_mid(self):
        upids = []
        ulb = self.bit_length
        while ulb:
            upid_type = self.bitbin.as_int(8)  # 1 byte
            ulb -= 8
            upid_length = self.bitbin.as_int(8)
            ulb -= 8
            upid_type_name, segmentation_upid = UpidDecoder(
                self.bitbin, upid_type, upid_length
            ).decode()
            mid_upid = {
                "upid_type": upid_type,
                "upid_type_name": upid_type_name,
                "upid_length": upid_length,
                "segmentation_upid": segmentation_upid,
            }
            ulb -= upid_length << 3
            upids.append(mid_upid)
        return upids

    def _decode_mpu(self):
        mpu_data = {
            "format_identifier": self.bitbin.as_int(32),
            "private_data": self.bitbin.as_hex(self.bit_length - 32),
        }
        return mpu_data

    def _decode_umid(self):
        chunks = []
        ulb = self.bit_length
        while ulb > 32:
            chunks.append(self.bitbin.as_hex(32)[2:])
            ulb -= 32
        return ".".join(chunks)

    def _decode_uri(self):
        return self.bitbin.as_charset(self.bit_length, charset)

    def _decode_no(self):
        return None

    def decode(self):
        """
        decode returns a upid determined by
        self.upid_type and upid_map below.
        """
        upid_map = {
            0x00: ["No UPID", self._decode_no],
            0x01: ["Deprecated", self._decode_uri],
            0x02: ["Deprecated", self._decode_uri],
            0x03: ["AdID", self._decode_uri],
            0x04: ["UMID", self._decode_umid],
            0x05: ["ISAN", self._decode_isan],
            0x06: ["ISAN", self._decode_isan],
            0x07: ["TID", self._decode_uri],
            0x08: ["AiringID", self._decode_air_id],
            0x09: ["ADI", self._decode_uri],
            0x10: ["UUID", self._decode_uri],
            0x11: ["SCR", self._decode_uri],
            0x0A: ["EIDR", self._decode_eidr],
            0x0B: ["ATSC", self._decode_atsc],
            0x0C: ["MPU", self._decode_mpu],
            0x0D: ["MID", self._decode_mid],
            0x0E: ["ADS Info", self._decode_uri],
            0x0F: ["URI", self._decode_uri],
            0xFD: ["Unknown", self._decode_uri],
        }
        if self.upid_type not in upid_map:
            self.upid_type = 0xFD
        if not self.upid_length:
            self.upid_type = 0x00
        upid_name = upid_map[self.upid_type][0]
        upid_value = upid_map[self.upid_type][1]()
        return upid_name, upid_value


def upid_encoder(nbin, upid_type, upid_length, seg_upid):
    """
    upid_encoder
    encodes segmentation_upid values to bites
    added to the nbin instance.

    Used by the SegmentationDescriptor class.
    """
    upid_map = {
        0x00: ["No UPID", _encode_no],
        0x01: ["Deprecated", _encode_uri],
        0x02: ["Deprecated", _encode_uri],
        0x03: ["AdID", _encode_uri],
        0x04: ["UMID", _encode_umid],
        0x07: ["TID", _encode_uri],
        0x0B: ["ATSC", _encode_atsc],
        0x09: ["ADI", _encode_uri],
        0x10: ["UUID", _encode_uri],
        0x11: ["SCR", _encode_uri],
        0x0A: ["EIDR", _encode_eidr],
        0x0C: ["MPU", _encode_mpu],
        0x0D: ["MID", _encode_mid],
        0x0E: ["ADS Info", _encode_uri],
        0x0F: ["URI", _encode_uri],
    }
    if upid_type in upid_map:
        upid_map[upid_type][1](nbin, seg_upid)

    upid_map_too = {
        0x05: ["ISAN", _encode_isan],
        0x06: ["ISAN", _encode_isan],
        0x08: ["AiringID", _encode_air_id],
    }
    if upid_type in upid_map_too:
        upid_map_too[upid_type][1](nbin, seg_upid, upid_length)


def _encode_air_id(nbin, seg_upid, upid_length):
    nbin.add_hex(seg_upid, (upid_length << 3))


def _encode_atsc(nbin, seg_upid):
    nbin.add_int(seg_upid["TSID"], 16)
    nbin.add_int(seg_upid["reserved"], 2)
    nbin.add_int(seg_upid["end_of_day"], 5)
    nbin.add_int(seg_upid["unique_for"], 9)
    nbin.add_bites(seg_upid["content_id"].encode("utf-8"))


def _encode_eidr(nbin, seg_upid):
    pre, post = seg_upid[3:].split("/", 1)
    nbin.add_int(int(pre), 16)
    nbin.add_hex(post.replace("-", ""), 80)


def _encode_isan(nbin, seg_upid, upid_length):
    nbin.add_hex(seg_upid, (upid_length << 3))


def _encode_mid(nbin, seg_upid):
    for mid_upid in seg_upid:
        nbin.add_int(mid_upid["upid_type"], 8)
        nbin.add_int(mid_upid["upid_length"], 8)
        upid_encoder(
            nbin,
            mid_upid["upid_type"],
            mid_upid["upid_length"],
            mid_upid["segmentation_upid"],
        )


def _encode_mpu(nbin, seg_upid):
    nbin.add_int(seg_upid["format_identifier"], 32)
    nbin.add_hex(seg_upid["private_data"])


def _encode_no(nbin, seg_upid):
    nbin.forward(0)


def _encode_umid(nbin, seg_upid):
    chunks = seg_upid.split(".")
    for chunk in chunks:
        nbin.add_hex(chunk, 32)


def _encode_uri(nbin, seg_upid):
    if len(seg_upid) > 0:
        seg_upid = seg_upid.encode()
        nbin.add_bites(seg_upid)



"""
segmentation.py

SCTE35 Segmentation Descriptor tables.
"""


table20 = {
    0x00: "Restrict Group 0",
    0x01: "Restrict Group 1",
    0x02: "Restrict Group 2",
    0x03: "No Restrictions",
}


table22 = {
    0x00: "Not Indicated",
    0x01: "Content Identification",
    0x10: "Program Start",
    0x11: "Program End",
    0x12: "Program Early Termination",
    0x13: "Program Breakaway",
    0x14: "Program Resumption",
    0x15: "Program Runover Planned",
    0x16: "Program RunoverUnplanned",
    0x17: "Program Overlap Start",
    0x18: "Program Blackout Override",
    0x19: "Program Start ??? In Progress",
    0x20: "Chapter Start",
    0x21: "Chapter End",
    0x22: "Break Start",
    0x23: "Break End",
    0x24: "Opening Credit Start",
    0x25: "Opening Credit End",
    0x26: "Closing Credit Start",
    0x27: "Closing Credit End",
    0x30: "Provider Advertisement Start",
    0x31: "Provider Advertisement End",
    0x32: "Distributor Advertisement Start",
    0x33: "Distributor Advertisement End",
    0x34: "Provider Placement Opportunity Start",
    0x35: "Provider Placement Opportunity End",
    0x36: "Distributor Placement Opportunity Start",
    0x37: "Distributor Placement Opportunity End",
    0x38: "Provider Overlay Placement Opportunity Start",
    0x39: "Provider Overlay Placement Opportunity End",
    0x3A: "Distributor Overlay Placement Opportunity Start",
    0x3B: "Distributor Overlay Placement Opportunity End",
    0x3C: "Provider Promo Start",
    0x3D: "Provider Promo End",
    0x3E: "Distributor Promo Start",
    0x3F: "Distributor Promo End",
    0x40: "Unscheduled Event Start",
    0x41: "Unscheduled Event End",
    0x42: "Alternate Content Opportunity Start",
    0x43: "Alternate Content Opportunity End",
    0x44: "Provider Ad Block Start",
    0x45: "Provider Ad Block End",
    0x46: "Distributor Ad Block Start",
    0x47: "Distributor Ad Block End",
    0x50: "Network Start",
    0x51: "Network End",
}



"""
SCTE35 Splice Descriptors
"""
from .bitn import BitBin
from .base import SCTE35Base
from .segmentation import table20, table22
from .upids import UpidDecoder, upid_encoder


def k_by_v(adict, avalue):
    """
    dict key lookup by value
    """
    for kay, vee in adict.items():
        if vee == avalue:
            return kay
    return None


class SpliceDescriptor(SCTE35Base):
    """
    SpliceDescriptor is the
    base class for all splice descriptors.
    It should not be used directly
    """

    def __init__(self, bites=None):
        self.tag = None
        self.descriptor_length = 0
        self.name = None
        self.identifier = "CUEI"
        self.bites = bites
        self.parse_tag_and_len()
        self.parse_id()
        self.provider_avail_id = None
        self.components = None

    def parse_tag_and_len(self):
        """
        parse_tag_and_len
        parses the descriptors tag and length
        from self.bites
        """
        if self.bites:
            self.tag = self.bites[0]
            self.descriptor_length = self.bites[1]
            self.bites = self.bites[2:]

    def parse_id(self):
        """
        parse splice descriptor identifier
        """
        if self.bites:
            self.identifier = self.bites[:4].decode()
            # disabled for ffmv30
            #      if self.identifier != "CUEI":
            #          raise Exception('Identifier Is Not "CUEI"')
            self.bites = self.bites[4:]

    def encode(self, nbin=None):
        """
        SpliceDescriptor.encode
        """
        nbin = self._chk_nbin(nbin)
        self._encode_id(nbin)
        return nbin

    def _encode_id(self, nbin):
        """
        parse splice descriptor identifier
        """
        # self.identifier = "CUEI"
        id_int = int.from_bytes(self.identifier.encode(), byteorder="big")
        nbin.add_int(id_int, 32)


class AudioDescriptor(SpliceDescriptor):
    """q
    Table 26 - audio_descriptor()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.tag = 4
        self.name = "Audio Descriptor"
        self.components = []
        self.audio_count = None

    def _decode_comp(self, bitbin):
        """
        AudioDescriptor._mk_comp parses
        for audio component data.
        """
        comp = {}
        comp["component_tag"] = bitbin.as_int(8)
        comp["ISO_code="] = bitbin.as_int(24)
        comp["bit_stream_mode"] = bitbin.as_int(3)
        comp["num_channels"] = bitbin.as_int(4)
        comp["full_srvc_audio"] = bitbin.as_flag(1)
        self.components.append(comp)

    @staticmethod
    def _encode_comp(comp, nbin):
        nbin.add_int(comp["component_tag"], 8)
        nbin.add_int(comp["ISO_code="], 24)
        nbin.add_int(comp["bit_stream_mode"], 3)
        nbin.add_int(comp["num_channels"], 4)
        nbin.add_flag(comp["full_srvc_audio"])

    def decode(self):
        """
        Decode SCTE35 Audio Descriptor
        """
        self.name = "Audio Descriptor"
        bitbin = BitBin(self.bites)
        a_c = bitbin.as_int(4)
        bitbin.forward(4)
        self.components = []
        while a_c:
            a_c -= 1
            self._decode_comp(bitbin)

    def encode(self, nbin=None):
        """
        encode SCTE35 Audio Descriptor
        """
        nbin = super().encode(nbin)
        nbin.add_int(self.audio_count, 4)
        nbin.forward(4)
        a_c = 0
        while a_c < self.audio_count:
            comp = self.components[a_c]
            self._encode_comp(comp, nbin)
            a_c += 1
        return nbin.bites


class AvailDescriptor(SpliceDescriptor):
    """
    Table 17 -  avail_descriptor()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.name = "Avail Descriptor"
        self.tag = 0

    def decode(self):
        """
        decode SCTE35 Avail Descriptor
        """
        bitbin = BitBin(self.bites)
        self.provider_avail_id = bitbin.as_int(32)

    def encode(self, nbin=None):
        """
        encode SCTE35 Avail Descriptor
        """
        nbin = super().encode(nbin)
        self._chk_var(int, nbin.add_int, "provider_avail_id", 32)
        return nbin.bites


class DtmfDescriptor(SpliceDescriptor):
    """
    Table 18 -  DTMF_descriptor()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.name = "DTMF Descriptor"
        self.preroll = None
        self.dtmf_count = 0
        self.dtmf_chars = None

    def decode(self):
        """
        decode SCTE35 Dtmf Descriptor
        """
        self.preroll = self.bites[0]
        self.dtmf_count = self.bites[1] >> 5
        self.bites = self.bites[2:]
        self.dtmf_chars = list(self.bites[: self.dtmf_count].decode())

    def encode(self, nbin=None):
        """
        encode SCTE35 Dtmf Descriptor
        """
        nbin = super().encode(nbin)
        self._chk_var(int, nbin.add_int, "preroll", 8)
        d_c = 0
        self._chk_var(int, nbin.add_int, "dtmf_count", 3)
        nbin.forward(5)
        while d_c < self.dtmf_count:
            nbin.add_int(ord(self.dtmf_chars[d_c]), 8)
            d_c += 1
        return nbin.bites


class TimeDescriptor(SpliceDescriptor):
    """
    Table 25 - time_descriptor()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.tag = 3
        self.name = "Time Descriptor"
        self.tai_seconds = 0
        self.tai_ns = 0
        self.utc_offset = 0

    def decode(self):
        """
        decode SCTE35 Time Descriptor
        """
        bitbin = BitBin(self.bites)
        self.tai_seconds = bitbin.as_int(48)
        self.tai_ns = bitbin.as_int(32)
        self.utc_offset = bitbin.as_int(16)

    def encode(self, nbin=None):
        """
        encode SCTE35 Time Descriptor
        """
        nbin = super().encode(nbin)
        self._chk_var(int, nbin.add_int, "tai_seconds", 48)
        self._chk_var(int, nbin.add_int, "tai_ns", 32)
        self._chk_var(int, nbin.add_int, "utc_offset", 16)
        return nbin.bites


class SegmentationDescriptor(SpliceDescriptor):
    """
    Table 19 - segmentation_descriptor()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.tag = 2
        self.name = "Segmentation Descriptor"
        self.segmentation_event_id = None
        self.segmentation_event_cancel_indicator = None
        self.component_count = None
        self.components = []
        self.program_segmentation_flag = None
        self.segmentation_duration_flag = None
        self.delivery_not_restricted_flag = None
        self.web_delivery_allowed_flag = None
        self.no_regional_blackout_flag = None
        self.archive_allowed_flag = None
        self.device_restrictions = None
        self.segmentation_duration = None
        self.segmentation_duration_ticks = None
        self.segmentation_message = None
        self.segmentation_upid_type = None
        self.segmentation_upid_type_name = None
        self.segmentation_upid_length = None
        self.segmentation_upid = None
        self.segmentation_type_id = None
        self.segment_num = None
        self.segments_expected = None
        self.sub_segment_num = None
        self.sub_segments_expected = None

    def decode(self):
        """
        decode a segmentation descriptor
        """
        bitbin = BitBin(self.bites)
        self.segmentation_event_id = bitbin.as_hex(32)  # 4 bytes
        self.segmentation_event_cancel_indicator = bitbin.as_flag(1)
        bitbin.forward(7)  # 1 byte
        if not self.segmentation_event_cancel_indicator:
            self._decode_flags(bitbin)  # 1 byte
            if not self.program_segmentation_flag:
                self._decode_components(bitbin)
            self._decode_segmentation(bitbin)

    def _decode_flags(self, bitbin):  # 1 byte for set flags
        self.program_segmentation_flag = bitbin.as_flag(1)
        self.segmentation_duration_flag = bitbin.as_flag(1)
        self.delivery_not_restricted_flag = bitbin.as_flag(1)
        if not self.delivery_not_restricted_flag:
            self.web_delivery_allowed_flag = bitbin.as_flag(1)
            self.no_regional_blackout_flag = bitbin.as_flag(1)
            self.archive_allowed_flag = bitbin.as_flag(1)
            self.device_restrictions = table20[bitbin.as_int(2)]
        else:
            bitbin.forward(5)

    def _decode_components(self, bitbin):
        self.component_count = c_c = bitbin.as_int(8)  # 1 byte
        while c_c:  # 6 bytes each
            c_c -= 1
            comp = {}
            comp["component_tag"] = bitbin.as_int(8)
            bitbin.forward(7)
            comp["pts_offset"] = bitbin.as_90k(33)
            self.components.append(comp)

    def _decode_segmentation(self, bitbin):
        if self.segmentation_duration_flag:
            self.segmentation_duration_ticks = bitbin.as_int(40)  # 5 bytes
            self.segmentation_duration = self.as_90k(self.segmentation_duration_ticks)
        self.segmentation_upid_type = bitbin.as_int(8)  # 1 byte
        self.segmentation_upid_length = bitbin.as_int(8)  # 1 byte
        self.segmentation_upid_type_name, self.segmentation_upid = UpidDecoder(
            bitbin, self.segmentation_upid_type, self.segmentation_upid_length
        ).decode()
        self.segmentation_type_id = bitbin.as_int(8)  # 1 byte
        if self.segmentation_type_id in table22:
            self.segmentation_message = table22[self.segmentation_type_id]
            self._decode_segments(bitbin)

    def _decode_segments(self, bitbin):
        self.segment_num = bitbin.as_int(8)  # 1 byte
        self.segments_expected = bitbin.as_int(8)  # 1 byte
        if self.segmentation_type_id in [0x34, 0x36, 0x38, 0x3A]:
            self.sub_segment_num = bitbin.as_int(8)  # 1 byte
            self.sub_segments_expected = bitbin.as_int(8)  # 1 byte

    def encode(self, nbin=None):
        """
        encode a segmentation descriptor
        """
        nbin = super().encode(nbin)
        self._chk_var(str, nbin.add_hex, "segmentation_event_id", 32)  # 4 bytes
        self._chk_var(bool, nbin.add_flag, "segmentation_event_cancel_indicator", 1)
        nbin.forward(7)  # 1 byte
        if not self.segmentation_event_cancel_indicator:
            self._encode_flags(nbin)  # 1 byte
            if not self.program_segmentation_flag:
                self._encode_components(nbin)
            self._encode_segmentation(nbin)
        return nbin.bites

    def _encode_components(self, nbin):
        nbin.add_int(self.component_count, 8)  # 1 byte
        c_c = 0
        while c_c < self.component_count:  # 6 bytes each
            comp = self.components[c_c]
            nbin.add_int(comp["component_tag"], 8)
            nbin.forward(7)
            nbin.add_90k(comp["pts_offset"], 33)
            c_c += 1

    def _encode_flags(self, nbin):  # 1 byte for set flags
        self._chk_var(bool, nbin.add_flag, "program_segmentation_flag", 1)
        self._chk_var(bool, nbin.add_flag, "segmentation_duration_flag", 1)
        self._chk_var(bool, nbin.add_flag, "delivery_not_restricted_flag", 1)
        if not self.delivery_not_restricted_flag:
            self._chk_var(bool, nbin.add_flag, "web_delivery_allowed_flag", 1)
            self._chk_var(bool, nbin.add_flag, "no_regional_blackout_flag", 1)
            self._chk_var(bool, nbin.add_flag, "archive_allowed_flag", 1)
            a_key = k_by_v(table20, self.device_restrictions)
            nbin.add_int(a_key, 2)
        else:
            nbin.reserve(5)

    def _encode_segmentation(self, nbin):
        if self.segmentation_duration_flag:
            if self.segmentation_duration_ticks and not self.segmentation_duration:
                self.segmentation_duration = self.as_90k(
                    self.segmentation_duration_ticks
                )
            self._chk_var(float, nbin.add_90k, "segmentation_duration", 40)  # 5 bytes
        self._chk_var(int, nbin.add_int, "segmentation_upid_type", 8)  # 1 byte
        self._chk_var(int, nbin.add_int, "segmentation_upid_length", 8)  # 1 byte
        upid_encoder(
            nbin,
            self.segmentation_upid_type,
            self.segmentation_upid_length,
            self.segmentation_upid,
        )
        self._chk_var(int, nbin.add_int, "segmentation_type_id", 8)  # 1 byte
        self._encode_segments(nbin)

    def _encode_segments(self, nbin):
        self._chk_var(int, nbin.add_int, "segment_num", 8)  # 1 byte
        self._chk_var(int, nbin.add_int, "segments_expected", 8)  # 1 byte
        if self.segmentation_type_id in [0x34, 0x36, 0x38, 0x3A]:
            self._chk_var(int, nbin.add_int, "sub_segment_num", 8)  # 1 byte
            self._chk_var(int, nbin.add_int, "sub_segments_expected", 8)  # 1 byte

    def _encode_segments(self, nbin):
        self._chk_var(int, nbin.add_int, "segment_num", 8)  # 1 byte
        self._chk_var(int, nbin.add_int, "segments_expected", 8)  # 1 byte
        if self.segmentation_type_id in [0x34, 0x36, 0x38, 0x3A]:
            self._chk_var(int, nbin.add_int, "sub_segment_num", 8)  # 1 byte
            self._chk_var(int, nbin.add_int, "sub_segments_expected", 8)  # 1 byte


# map of known descriptors and associated classes
descriptor_map = {
    0: AvailDescriptor,
    1: DtmfDescriptor,
    2: SegmentationDescriptor,
    3: TimeDescriptor,
    4: AudioDescriptor,
}


def splice_descriptor(bites):
    """
    splice_descriptor reads the
    descriptor tag and decodes and
    returns an instance self._descriptor_map[tag]
    """
    tag = bites[0]
    if tag not in descriptor_map:
        return False
    spliced = descriptor_map[tag](bites)
    spliced.decode()
    return spliced



#!/usr/bin/env python3

import setuptools
import threefive

with open("README.md", "r") as fh:
    readme = fh.read()

setuptools.setup(
    name="threefive",
    version=threefive.version(),
    author="Adrian",
    author_email="spam@iodisco.com",
    description="Southern Fried SCTE35",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/futzu/threefive",
    install_requires=[
        "new_reader >= 0.0.9",
        "pyaes",
    ],
    scripts=["bin/threefive"],
    packages=setuptools.find_packages(),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    python_requires=">=3.6",
)



"""
threefive.base contains
the class SCTE35Base.
"""
import json
from .bitn import NBin


class SCTE35Base:
    """
    SCTE35Base is a base class for
    SpliceCommand and SpliceDescriptor classes
    """

    ROLLOVER = 8589934591

    def __repr__(self):
        return str(self.__dict__)

    @staticmethod
    def as_90k(int_time):
        return round((int_time / 90000.0), 6)

    @staticmethod
    def as_hms(secs_of_time):
        """
        as_hms converts timestamp to
        00:00:00.000 format
        """
        hours, seconds = divmod(secs_of_time, 3600)
        mins, seconds = divmod(seconds, 60)
        seconds = round(seconds, 3)
        output = f"{int(hours):02}:{int(mins):02}:{seconds:02}"
        if len(output.split(".")[1]) < 2:
            output += "0"
        return output

    def get(self):
        """
        Returns instance as a dict
        """
        return self.kv_clean()

    def get_json(self):
        """
        get_json returns the instance
        data as json.
        """
        return json.dumps(self.get(), indent=4)

    def kv_clean(self):
        """
        kv_clean removes items from a dict if the value is None
        """

        def b2l(val):
            if isinstance(val, (SCTE35Base)):
                val.kv_clean()
            if isinstance(val, (list)):
                val = [b2l(v) for v in val]
            if isinstance(val, (dict)):
                val = {k: b2l(v) for k, v in val.items()}
            if isinstance(val, (bytes, bytearray)):
                val = list(val)
            return val

        return {k: b2l(v) for k, v in vars(self).items() if v is not None}

    @staticmethod
    def _chk_nbin(nbin):
        if not nbin:
            nbin = NBin()
        return nbin

    def load(self, stuff):
        """
        load is used to load
        data from a dict or json string
        """
        if isinstance(stuff, str):
            stuff = json.loads(stuff)
        if isinstance(stuff, dict):
            self.__dict__.update(stuff)

    def _chk_var(self, var_type, nbin_method, var_name, bit_count):
        """
        _chk_var is used to check var values and types before encoding
        """
        var_value = self.__dict__[var_name]
        if var_value is None:
            err_mesg = (
                f"\033[7m{var_name} is not set, it should be type {var_type}\033[27m"
            )
            raise ValueError(err_mesg)
        if not isinstance(var_value, var_type):
            err_mesg = f' \033[7m{var_name} is "{var_value}", it should be type {var_type}\033[27m '
            raise ValueError(err_mesg)
        nbin_method(var_value, bit_count)



"""
section.py

SCTE35 Splice Info Section
"""
from .bitn import BitBin
from .base import SCTE35Base

sap_map = {
    "0x0": "Type 1 Closed GOP with no leading pictures",
    "0x1": "Type 2 Closed GOP with leading pictures",
    "0x2": "Type 3 Open GOP",
    "0x3": "No Sap Type",
}


class SpliceInfoSection(SCTE35Base):
    """
    The SCTE-35 splice info section
    data.
    """

    def __init__(self):
        self.table_id = None
        self.section_syntax_indicator = None
        self.private = None
        # sap_type used to be marked reserved.
        self.sap_type = None
        self.sap_details = None
        self.section_length = None
        self.protocol_version = None
        self.encrypted_packet = None
        self.encryption_algorithm = None
        self.pts_adjustment_ticks = None
        self.pts_adjustment = None
        self.cw_index = None
        self.tier = None
        self.splice_command_length = None
        self.splice_command_type = None
        self.descriptor_loop_length = 0
        self.crc = None

    def decode(self, bites):
        """
        InfoSection.decode
        """
        bitbin = BitBin(bites)
        self.table_id = bitbin.as_hex(8)
        if self.table_id != "0xfc":
            raise ValueError(
                ("splice_info_section.table_id should be 0xfc Not: ", self.table_id)
            )
        self.section_syntax_indicator = bitbin.as_flag(1)
        self.private = bitbin.as_flag(1)
        self.sap_type = bitbin.as_hex(2)
        self.sap_details = sap_map[self.sap_type]
        self.section_length = bitbin.as_int(12)
        self.protocol_version = bitbin.as_int(8)
        self.encrypted_packet = bitbin.as_flag(1)
        self.encryption_algorithm = bitbin.as_int(6)
        self.pts_adjustment_ticks = bitbin.as_int(33)
        self.pts_adjustment = self.as_90k(self.pts_adjustment_ticks)
        self.cw_index = bitbin.as_hex(8)
        self.tier = bitbin.as_hex(12)
        self.splice_command_length = bitbin.as_int(12)
        self.splice_command_type = bitbin.as_int(8)

    def _encode_table_id(self, nbin):
        """
        encode SpliceInfoSection.table_id
        """
        self.table_id = "0xfc"
        nbin.add_hex(self.table_id, 8)

    def _encode_section_syntax_indicator(self, nbin):
        """
        encode SpliceInfoSection.section_syntax_indicator
        """
        self.section_syntax_indicator = False
        nbin.add_flag(self.section_syntax_indicator, 1)

    def _encode_private_flag(self, nbin):
        """
        encode SpliceInfoSection.private
        """
        self.private = False
        nbin.add_flag(self.private, 1)

    def _encode_sap(self, nbin):
        """
        the 3 reserved bits now map SAP
        """
        if self.sap_type not in sap_map:
            self.sap_type = "0x3"
        self.sap_details = sap_map[self.sap_type]
        nbin.add_hex(self.sap_type, 2)

    def _encode_section_length(self, nbin):
        """
        encode SpliceInfoSection.section_length
        """
        if not self.section_length:
            self.section_length = 11
        nbin.add_int(self.section_length, 12)

    def _encode_protocol_version(self, nbin):
        """
        encode SpliceInfoSection.protocol_version
        """
        if not self.protocol_version:
            self.protocol_version = 0
        nbin.add_int(self.protocol_version, 8)

    def _encode_encrypted(self, nbin):
        """
        encode SpliceInfoSection.encrypted_packet
        and SpliceInfoSection.encyption_algorithm
        """
        if not self.encrypted_packet:
            self.encrypted_packet = False
        nbin.add_flag(self.encrypted_packet)
        if not self.encryption_algorithm:
            self.encryption_algorithm = 0
        nbin.add_int(self.encryption_algorithm, 6)

    def _encode_pts_adjustment(self, nbin):
        """
        encode SpliceInfoSection.pts_adjustment
        """
        if not self.pts_adjustment:
            self.pts_adjustment = 0.0
        nbin.add_90k(self.pts_adjustment, 33)

    def _encode_cw_index(self, nbin):
        """
        encode SpliceInfoSection.cw_index
        """
        if not self.cw_index:
            self.cw_index = "0x0"
        nbin.add_hex(self.cw_index, 8)

    def _encode_tier(self, nbin):
        """
        encode SpliceInfoSection.tier
        """
        if not self.tier:
            self.tier = "0xfff"
        nbin.add_hex(self.tier, 12)

    def _encode_splice_command(self, nbin):
        """
        encode Splice.InfoSection.splice_command_length
        and Splice.InfoSection.splice_command_type
        """
        if not self.splice_command_length:
            self.splice_command_length = 0
        nbin.add_int(self.splice_command_length, 12)
        if not self.splice_command_type:
            self.splice_command_type = 0
        nbin.add_int(self.splice_command_type, 8)

    def encode(self, nbin=None):
        """
        SpliceInfoSection.encode
        takes the vars from an instance and
        encodes them as bytes.
        """
        nbin = self._chk_nbin(nbin)
        self._encode_table_id(nbin)
        self._encode_section_syntax_indicator(nbin)
        self._encode_private_flag(nbin)
        self._encode_sap(nbin)
        self._encode_section_length(nbin)
        self._encode_protocol_version(nbin)
        self._encode_encrypted(nbin)
        self._encode_pts_adjustment(nbin)
        self._encode_cw_index(nbin)
        self._encode_tier(nbin)
        self._encode_splice_command(nbin)
        return nbin.bites



"""
crc.py  crc32 function for encoding.

"""

POLY = 0x104C11DB7
INIT_VALUE = 0xFFFFFFFF
SIZE_BITS = 32


def _bytecrc(crc, poly, num):
    mask = 1 << (num - 1)
    i = 8
    while i:
        crc = (crc << 1, crc << 1 ^ poly)[crc & mask != 0]
        i -= 1
    mask = (1 << num) - 1
    crc = crc & mask
    return crc


def _mk_table():
    mask = (1 << SIZE_BITS) - 1
    poly = POLY & mask
    return [_bytecrc(i << (SIZE_BITS - 8), poly, SIZE_BITS) for i in range(256)]


def crc32(data):
    table = _mk_table()
    emvee = memoryview(data)
    crc = INIT_VALUE & 0xFFFFFFFF
    for bite in emvee.tobytes():
        crc = table[bite ^ ((crc >> 24) & 0xFF)] ^ ((crc << 8) & 0xFFFFFF00)
    return crc



"""
SCTE35 Splice Commands
"""
from .bitn import BitBin
from .base import SCTE35Base


class SpliceCommand(SCTE35Base):
    """
    Base class, not used directly.
    """

    def __init__(self, bites=None):
        self.command_length = 0
        self.command_type = None
        self.name = None
        self.bites = bites

    def decode(self):
        """
        default decode method
        """

    def _set_len(self, start, end):
        """
        _set_len sets
        self.command_length
        """
        self.command_length = (start - end) >> 3

    def encode(self, nbin=None):
        """
        encode
        """
        nbin = self._chk_nbin(nbin)
        return nbin.bites


class BandwidthReservation(SpliceCommand):
    """
    Table 11 - bandwidth_reservation()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.command_type = 7
        self.name = "Bandwidth Reservation"

    def decode(self):
        """
        BandwidthReservation.decode method
        """


class PrivateCommand(SpliceCommand):
    """
    Table 12 - private_command
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.command_type = 255
        self.name = "Private Command"
        self.identifier = None

    def decode(self):
        """
        PrivateCommand.decode method
        """
        self.identifier = int.from_bytes(
            self.bites[0:3], byteorder="big"
        )  # 3 bytes = 24 bits
        self.bites = self.bites[3:]

    def encode(self, nbin=None):
        """
        encode private command
        """
        nbin = self._chk_nbin(nbin)
        self._chk_var(int, nbin.add_int, "identifier", 24)  # 3 bytes = 24 bits
        return nbin.bites


class SpliceNull(SpliceCommand):
    """
    Table 7 - splice_null()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.command_type = 0
        self.name = "Splice Null"


class TimeSignal(SpliceCommand):
    """
    Table 10 - time_signal()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.command_type = 6
        self.name = "Time Signal"
        self.time_specified_flag = None
        self.pts_time = None
        self.pts_time_ticks = None

    def decode(self):
        """
        TimeSignal.decode method
        """
        bitbin = BitBin(self.bites)
        start = bitbin.idx
        self._splice_time(bitbin)
        self._set_len(start, bitbin.idx)

    def encode(self, nbin=None):
        """
        encode converts TimeSignal vars
        to bytes
        """
        nbin = self._chk_nbin(nbin)
        self._encode_splice_time(nbin)
        return nbin.bites

    def _splice_time(self, bitbin):
        """
        parse_pts is called by either a
        TimeSignal or SpliceInsert instance
        to decode pts.
        """
        self.time_specified_flag = bitbin.as_flag(1)
        if self.time_specified_flag:
            bitbin.forward(6)
            self.pts_time_ticks = bitbin.as_int(33)
            self.pts_time = self.as_90k(self.pts_time_ticks)
        else:
            bitbin.forward(7)

    def _encode_splice_time(self, nbin):
        self._chk_var(bool, nbin.add_flag, "time_specified_flag", 1)
        if self.time_specified_flag:
            nbin.reserve(6)
            if self.pts_time_ticks and not self.pts_time:
                self.pts_time = self.as_90k(self.pts_time_ticks)
            self._chk_var(float, nbin.add_90k, "pts_time", 33)
        else:
            nbin.reserve(7)


class SpliceInsert(TimeSignal):
    """
    Table 9 - splice_insert()
    """

    def __init__(self, bites=None):
        super().__init__(bites)
        self.command_type = 5
        self.name = "Splice Insert"
        self.break_auto_return = None
        self.break_duration = None
        self.break_duration_ticks = None
        self.splice_event_id = None
        self.splice_event_cancel_indicator = None
        self.out_of_network_indicator = None
        self.program_splice_flag = None
        self.duration_flag = None
        self.splice_immediate_flag = None
        self.component_count = None
        self.components = None
        self.unique_program_id = None
        self.avail_num = None
        self.avail_expected = None

    def decode(self):
        """
        SpliceInsert.decode
        """
        bitbin = BitBin(self.bites)
        start = bitbin.idx
        self._decode_event(bitbin)
        if not self.splice_event_cancel_indicator:
            self._decode_flags(bitbin)
            if self.program_splice_flag:
                if not self.splice_immediate_flag:
                    self._splice_time(bitbin)
            else:
                self._decode_components(bitbin)
                if not self.splice_immediate_flag:
                    self._splice_time(bitbin)
            self._decode_break(bitbin)
            self._decode_unique_avail(bitbin)
        self._set_len(start, bitbin.idx)

    def _decode_break(self, bitbin):
        """
        SpliceInsert._decode_break(bitbin) is called
        if SpliceInsert.duration_flag is set
        """
        if self.duration_flag:
            self.break_auto_return = bitbin.as_flag(1)
            bitbin.forward(6)
            self.break_duration_ticks = bitbin.as_int(33)
            self.break_duration = self.as_90k(self.break_duration_ticks)

    def _decode_components(self, bitbin):
        """
        SpliceInsert._decode_components loops
        over SpliceInsert.components,
        and is called from SpliceInsert.decode()
        """
        self.component_count = bitbin.as_int(8)
        self.components = []
        for i in range(0, self.component_count):
            self.components[i] = bitbin.as_int(8)

    def _decode_event(self, bitbin):
        """
        SpliceInsert._decode_event parses
        self.splice_event_id and self.splice_event_cancel_indicator
        and is called from SpliceInsert.decode()
        """
        self.splice_event_id = bitbin.as_int(32)
        self.splice_event_cancel_indicator = bitbin.as_flag(1)
        bitbin.forward(7)

    def _decode_flags(self, bitbin):
        """
        SpliceInsert._decode_flags parses four bit flags
        """
        self.out_of_network_indicator = bitbin.as_flag(1)
        self.program_splice_flag = bitbin.as_flag(1)
        self.duration_flag = bitbin.as_flag(1)
        self.splice_immediate_flag = bitbin.as_flag(1)
        bitbin.forward(4)

    def _decode_unique_avail(self, bitbin):
        self.unique_program_id = bitbin.as_int(16)
        self.avail_num = bitbin.as_int(8)
        self.avail_expected = bitbin.as_int(8)

    def encode(self, nbin=None):
        """
        SpliceInsert.encode
        """
        nbin = self._chk_nbin(nbin)
        self._encode_event(nbin)
        if not self.splice_event_cancel_indicator:
            self._encode_flags(nbin)
            if self.program_splice_flag:
                if not self.splice_immediate_flag:
                    self._encode_splice_time(nbin)
            else:
                self._encode_components(nbin)
                if not self.splice_immediate_flag:
                    self._encode_splice_time(nbin)
            self._encode_break(nbin)
            self._encode_unique_avail(nbin)
        return nbin.bites

    def _encode_event(self, nbin):
        """
        SpliceInsert._encode_event encodes
        self.splice_event_id and self.splice_event_cancel_indicator
        and is called from SpliceInsert.encode()
        """
        self._chk_var(int, nbin.add_int, "splice_event_id", 32)
        self._chk_var(bool, nbin.add_flag, "splice_event_cancel_indicator", 1)
        nbin.forward(7)

    def _encode_break(self, nbin):
        """
        SpliceInsert._encode_break(nbin) is called
        if SpliceInsert.duration_flag is set
        """
        if self.duration_flag:
            self._chk_var(bool, nbin.add_flag, "break_auto_return", 1)
            nbin.forward(6)
            self._chk_var(float, nbin.add_90k, "break_duration", 33)

    def _encode_flags(self, nbin):
        """
        SpliceInsert._encode_flags converts four flags
        to bits
        """
        self._chk_var(bool, nbin.add_flag, "out_of_network_indicator", 1)
        self._chk_var(bool, nbin.add_flag, "program_splice_flag", 1)
        self._chk_var(bool, nbin.add_flag, "duration_flag", 1)
        self._chk_var(bool, nbin.add_flag, "splice_immediate_flag", 1)
        nbin.forward(4)

    def _encode_components(self, nbin):
        """
        SpliceInsert._encode_components loops
        over SpliceInsert.components,
        and is called from SpliceInsert.encode()
        """
        nbin.add_int(self.component_count, 8)
        for i in range(0, self.component_count):
            nbin.add_int(self.components[i], 8)

    def _encode_unique_avail(self, nbin):
        self._chk_var(int, nbin.add_int, "unique_program_id", 16)
        self._chk_var(int, nbin.add_int, "avail_num", 8)
        self._chk_var(int, nbin.add_int, "avail_expected", 8)


class SpliceSchedule(SpliceCommand):
    """
    Table 8 - splice_schedule()
    """

    class SpliceEvent(SpliceInsert):
        """
        SpliceEvent is a class declared in
        the SpliceSchedule class.
        """

        def __init__(self):
            """
            SpliceEvent
            """
            super().__init__(None)
            self.name = None
            self.utc_splice_time = None

        def _decode_components(self, bitbin):
            """
            SpliceEvent._decode_components
            """
            component_count = bitbin.as_int(8)
            self.components = []
            for j in range(0, component_count):
                self.components[j] = {
                    "component_tag": bitbin.as_int(8),
                    "utc_splice_time": bitbin.as_int(32),
                }

        def decode(self, bitbin):
            """
            SpliceEvent.decode
            """
            self._decode_event(bitbin)
            if not self.splice_event_cancel_indicator:
                self._decode_flags(bitbin)
                bitbin.forward(5)
                if self.program_splice_flag:
                    self.utc_splice_time = bitbin.as_int(32)
                else:
                    self._decode_components(bitbin)
                self._decode_break(bitbin)
                self._decode_unique_avail(bitbin)

    def __init__(self, bites=None):
        """
        SpliceSchedule.__init__
        """
        super().__init__(bites)
        self.command_type = 4
        self.name = "Splice Schedule"
        self.splices = []

    def decode(self):
        """
        SpliceSchedule.decode
        """
        bitbin = BitBin(self.bites)
        start = bitbin.idx
        splice_count = bitbin.as_int(8)
        for i in range(0, splice_count):
            self.splices[i] = self.SpliceEvent()
            self.splices[i].decode(bitbin)
        self._set_len(start, bitbin.idx)


command_map = {
    0: SpliceNull,
    4: SpliceSchedule,
    5: SpliceInsert,
    6: TimeSignal,
    7: BandwidthReservation,
    255: PrivateCommand,
}



"""
threeefive.version

Odd number versions are releases.
Even number versions are testing builds between releases.

Used to set version in setup.py
and as an easy way to check which
version you have installed.
"""

MAJOR = "2"
MINOR = "3"
MAINTAINENCE = "57"


def version():
    """
    version prints threefives version as a string
    """
    return f"{MAJOR}.{MINOR}.{MAINTAINENCE}"


def version_number():
    """
    version_number returns version as an int.
    if version() returns 2.3.01
    version_number will return 2301
    """
    return int(f"{MAJOR}{MINOR}{MAINTAINENCE}")
